use std::sync::Arc;

use base64::prelude::*;

use chrono::serde::ts_seconds;
use chrono::{DateTime, Utc};
use cookie::time::{Duration, OffsetDateTime};
use jwt::{SignWithKey, VerifyWithKey};
use openidconnect::{
    core::{CoreIdTokenVerifier, CoreResponseType},
    AuthenticationFlow, AuthorizationCode, CsrfToken, Nonce, Scope,
};
use rand::{distr::Alphanumeric, rng, Rng};
use warp::{
    http::{header, Response, StatusCode},
    reject::Rejection,
    reply::Reply,
};

use crate::{AuthError, Server};

pub const COOKIE_AUTH_TOKEN: &str = "knotes_auth_token";
pub const COOKIE_AUTH_NONCE: &str = "knotes_auth_nonce";
pub const COOKIE_AUTH_CSRF: &str = "knotes_auth_csrf";

#[derive(Debug, serde::Serialize, serde::Deserialize)]
pub struct UserToken {
    pub display_name: String,
    pub email: String,

    // Unused for now, might be checked later do allow some basic session management.
    // Issued At Time
    #[allow(dead_code)]
    #[serde(with = "ts_seconds")]
    pub iat: DateTime<Utc>,
    // JWT Id
    #[allow(dead_code)]
    pub jti: String,
}

#[derive(Debug, serde::Serialize, serde::Deserialize)]
struct AuthState {
    redirect_url: Option<String>,
    csrf_token: String,
}

#[derive(Debug, serde::Deserialize)]
pub struct AuthLoginParams {
    redirect_url: Option<String>,
}

pub async fn handle_auth_login(
    server: Arc<Server>,
    params: AuthLoginParams,
) -> Result<impl Reply, Rejection> {
    let csrf_token = CsrfToken::new_random();
    let state = AuthState {
        redirect_url: params.redirect_url,
        csrf_token: csrf_token.clone().into_secret(),
    };
    let state = CsrfToken::new(
        BASE64_URL_SAFE.encode(
            state
                .sign_with_key(&server.signing_key)
                .map_err(|_| warp::reject::custom(AuthError))?,
        ),
    );
    // From what I understand I don't need the csrf_token
    let (auth_url, _csrf_token, nonce) = server
        .openid_client
        .authorize_url(
            AuthenticationFlow::<CoreResponseType>::AuthorizationCode,
            || state,
            Nonce::new_random,
        )
        .add_scope(Scope::new("openid".to_string()))
        .add_scope(Scope::new("profile".to_string()))
        .add_scope(Scope::new("email".to_string()))
        .url();

    Ok(Response::builder()
        .header(
            header::SET_COOKIE,
            cookie::Cookie::build((COOKIE_AUTH_NONCE, nonce.secret()))
                .path("/")
                .http_only(true)
                .build()
                .to_string(),
        )
        .header(
            header::SET_COOKIE,
            cookie::Cookie::build((COOKIE_AUTH_CSRF, csrf_token.secret()))
                .path("/")
                .http_only(true)
                .build()
                .to_string(),
        )
        .header(header::LOCATION, auth_url.as_str())
        .status(StatusCode::FOUND)
        .body("Login")
        .unwrap())
}

#[derive(Debug, serde::Deserialize)]
pub struct AuthCallbackParams {
    code: String,
    state: String,
}

pub async fn handle_auth_callback(
    server: Arc<Server>,
    nonce: String,
    csrf_token: String,
    params: AuthCallbackParams,
) -> Result<impl Reply, Rejection> {
    let code = AuthorizationCode::new(params.code);
    let nonce = Nonce::new(nonce);
    let state: AuthState = String::from_utf8(
        BASE64_URL_SAFE
            .decode(params.state)
            .map_err(|_| warp::reject::custom(AuthError))?,
    )
    .map_err(|_| warp::reject::custom(AuthError))?
    .verify_with_key(&server.signing_key)
    .map_err(|_| warp::reject::custom(AuthError))?;

    if state.csrf_token != csrf_token {
        return Err(warp::reject::custom(AuthError));
    }

    let token_response = server
        .openid_client
        .exchange_code(code)
        .map_err(|_| warp::reject::custom(AuthError))?
        .request_async(&server.http_client)
        .await
        .map_err(|_| warp::reject::custom(AuthError))?;

    let id_token_verifier: CoreIdTokenVerifier = server.openid_client.id_token_verifier();
    let id_token_claims = token_response
        .extra_fields()
        .id_token()
        .ok_or(warp::reject::custom(AuthError))?
        .claims(&id_token_verifier, &nonce)
        .map_err(|_| warp::reject::custom(AuthError))?;

    let token_id: String = rng()
        .sample_iter(&Alphanumeric)
        .take(32)
        .map(char::from)
        .collect();

    let display_name = id_token_claims
        .given_name()
        .ok_or(warp::reject::custom(AuthError))?
        .get(None)
        .ok_or(warp::reject::custom(AuthError))?
        .to_string();
    let email = id_token_claims
        .email()
        .ok_or(warp::reject::custom(AuthError))?
        .to_string();

    println!("[INFO] Authenticated: {} <{}>", display_name, email);

    let user_token = UserToken {
        display_name,
        email,
        iat: Utc::now(),
        jti: token_id,
    };
    let user_token = user_token
        .sign_with_key(&server.signing_key)
        .map_err(|_| warp::reject::custom(AuthError))?;

    Ok(Response::builder()
        .header(
            header::SET_COOKIE,
            cookie::Cookie::build((COOKIE_AUTH_NONCE, ""))
                .path("/")
                .http_only(true)
                .build()
                .to_string(),
        )
        .header(
            header::SET_COOKIE,
            cookie::Cookie::build((COOKIE_AUTH_CSRF, ""))
                .expires(OffsetDateTime::UNIX_EPOCH)
                .path("/")
                .http_only(true)
                .build()
                .to_string(),
        )
        .header(
            header::SET_COOKIE,
            cookie::Cookie::build((COOKIE_AUTH_TOKEN, user_token))
                .expires(OffsetDateTime::now_utc() + Duration::days(365))
                .path("/")
                .build()
                .to_string(),
        )
        .header(
            header::LOCATION,
            state.redirect_url.unwrap_or("/app".to_string()),
        )
        .status(StatusCode::FOUND)
        .body("")
        .unwrap())
}
