pub mod auth;

use auth::{
    handle_auth_callback, handle_auth_login, UserToken, COOKIE_AUTH_CSRF, COOKIE_AUTH_NONCE,
    COOKIE_AUTH_TOKEN,
};
use dotenvy::dotenv;
use futures_util::StreamExt;
use hmac::{Hmac, Mac};
use jwt::VerifyWithKey;
use openidconnect::{
    core::{
        CoreAuthDisplay, CoreAuthPrompt, CoreClient, CoreErrorResponseType, CoreGenderClaim,
        CoreJsonWebKey, CoreJweContentEncryptionAlgorithm, CoreProviderMetadata,
        CoreRevocableToken, CoreRevocationErrorResponse, CoreTokenIntrospectionResponse,
        CoreTokenResponse,
    },
    reqwest, Client, ClientId, ClientSecret, EmptyAdditionalClaims, EndpointMaybeSet,
    EndpointNotSet, EndpointSet, IssuerUrl, RedirectUrl, StandardErrorResponse,
};
use rocksdb::TransactionDB;
use sha2::Sha256;
use std::fs;
use std::{
    collections::HashMap,
    env,
    path::PathBuf,
    str::FromStr,
    sync::{Arc, Weak},
};
use tokio::sync::{Mutex, RwLock};
use warp::{
    reject::Reject,
    ws::{WebSocket, Ws},
    Filter, Rejection, Reply,
};
use yrs::{sync::Awareness, types::ToJson, Array, Doc, ReadTxn, Transact, WriteTxn};
use yrs_kvstore::DocOps;
use yrs_rocksdb::RocksDBStore;
use yrs_warp::{
    broadcast::BroadcastGroup,
    ws::{WarpSink, WarpStream},
};

// Web socket flow:
// -> Handle connection (ws_handler)
// -> Check permissions (Not doing for new)
// -> Get or load document
// ->
//
//

type OpenidClient = Client<
    EmptyAdditionalClaims,
    CoreAuthDisplay,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJsonWebKey,
    CoreAuthPrompt,
    StandardErrorResponse<CoreErrorResponseType>,
    CoreTokenResponse,
    CoreTokenIntrospectionResponse,
    CoreRevocableToken,
    CoreRevocationErrorResponse,
    EndpointSet,
    EndpointNotSet,
    EndpointNotSet,
    EndpointNotSet,
    EndpointMaybeSet,
    EndpointMaybeSet,
>;

#[derive(Debug)]
struct AuthError;
impl Reject for AuthError {}

pub struct Connection {
    bcast: BroadcastGroup,
    db: Arc<TransactionDB>,
    #[allow(dyn_drop)]
    _db_sub: Arc<dyn Drop + Send + Sync>,
}

pub struct Server {
    // There is something to be said for keeping the broadcast group in memory for a bit after all
    // clients disconnect, but for now we don't bother.
    pub open_docs: RwLock<HashMap<String, Weak<Connection>>>,

    pub http_client: reqwest::Client,
    pub openid_client: OpenidClient,

    pub signing_key: Hmac<Sha256>,

    pub data_dir: PathBuf,
}

impl Server {
    pub fn new(
        http_client: reqwest::Client,
        openid_client: OpenidClient,
        signing_key: Hmac<Sha256>,
        data_dir: PathBuf,
    ) -> Self {
        Self {
            open_docs: RwLock::default(),
            http_client,
            openid_client,
            signing_key,
            data_dir,
        }
    }
    pub async fn get_or_create_doc(&self, name: String) -> Arc<Connection> {
        let open_docs = self.open_docs.read().await;
        match open_docs.get(&name).and_then(Weak::upgrade) {
            Some(connection) => connection.clone(),
            None => {
                drop(open_docs);
                let mut open_docs = self.open_docs.write().await;
                let doc = Doc::new();

                let data_dir = self.data_dir.join(name.clone());
                if let Err(e) = fs::create_dir_all(&data_dir) {
                    panic!("Was unable to create data directory for note {}, due to the following error. Something is very wrong!\n{}", name, e)
                }

                let db =
                    Arc::new(TransactionDB::open_default(data_dir).expect("Failed to open DB"));

                // Subscribe the DB to updates
                let sub = {
                    let db = db.clone();
                    let name = name.clone();
                    doc.observe_update_v1(move |doc_txn, e| {
                        let txn = RocksDBStore::from(db.transaction());
                        let i = txn.push_update(&name, &e.update).unwrap();
                        if i % 128 == 0 {
                            // compact updates into document
                            txn.flush_doc(&name).unwrap();
                        }
                        txn.commit().unwrap();
                    })
                    .unwrap()
                };

                {
                    // Load document from DB
                    let mut txn = doc.transact_mut();
                    let db_txn = RocksDBStore::from(db.transaction());
                    db_txn.load_doc(&name, &mut txn).unwrap();
                }

                let awareness = Arc::new(RwLock::new(Awareness::new(doc)));
                let group = BroadcastGroup::new(awareness, 32).await;
                let connection = Arc::new(Connection {
                    bcast: group,
                    db,
                    _db_sub: sub,
                });

                open_docs.insert(name, Arc::downgrade(&connection));

                connection
            }
        }
    }
}

async fn create_openid_client(
    http_client: &reqwest::Client,
    issuer_url: IssuerUrl,
    client_id: ClientId,
    client_secret: ClientSecret,
    redirect_url: RedirectUrl,
) -> OpenidClient {
    let provider_metadata = CoreProviderMetadata::discover_async(issuer_url, http_client)
        .await
        .expect("Failed to discover OpenID Connect provider metadata");

    CoreClient::from_provider_metadata(provider_metadata, client_id, Some(client_secret))
        .set_redirect_uri(redirect_url)
}

#[tokio::main]
async fn main() {
    // Allow loading .env to fail, since it won't be available in docker.
    let _ = dotenv();

    let data_dir = env::var("DATA_DIR").expect("DATA_DIR not set");
    let data_dir = PathBuf::from_str(&data_dir).expect("DATA_DIR is not a valid path");
    let frontend_dir = env::var("FRONTEND_DIR").expect("FRONTEND_DIR not set");
    let frontend_dir = PathBuf::from_str(&frontend_dir).expect("FRONTEND_DIR is not a valid path");
    let index_file = frontend_dir.join("index.html");

    let signing_key = env::var("AUTH_SECRET").expect("AUTH_SECRET not set");
    let signing_key =
        Hmac::new_from_slice(signing_key.as_bytes()).expect("AUTH_SECRET is not a valid HMAC key");

    let app_url = env::var("APP_URL").expect("APP_URL not set");

    let http_client = reqwest::ClientBuilder::new()
        // Following redirects opens the client up to SSRF vulnerabilities.
        .redirect(reqwest::redirect::Policy::none())
        .build()
        .expect("Failed to create HTTP client");

    let openid_client = create_openid_client(
        &http_client,
        IssuerUrl::new(env::var("AUTH_ISSUER_URL").expect("AUTH_ISSUER_URL not set"))
            .expect("Issuer URL invalid"),
        ClientId::new(env::var("AUTH_CLIENT_ID").expect("AUTH_CLIENT_ID not set")),
        ClientSecret::new(env::var("AUTH_CLIENT_SECRET").expect("AUTH_CLIENT_SECRET not set")),
        RedirectUrl::new(format!("{}/api/auth/callback", app_url))
            .expect("Redirect URL is invalid (app url?)"),
    )
    .await;

    let server = Arc::new(Server::new(
        http_client,
        openid_client,
        signing_key,
        data_dir,
    ));

    let ws = {
        let server = server.clone();
        warp::path("sync")
            .and(warp::path::param())
            .and(warp::ws())
            .and(warp::any().map(move || server.clone()))
            .and(warp::cookie(COOKIE_AUTH_TOKEN))
            .and_then(ws_handler)
    };

    let api_routes = {
        let login_route = {
            let server = server.clone();
            warp::path!("api" / "auth" / "login")
                .map(move || server.clone())
                .and(warp::query())
                .and_then(handle_auth_login)
        };

        let callback_route = {
            let server = server.clone();
            warp::path!("api" / "auth" / "callback")
                .map(move || server.clone())
                .and(warp::cookie(COOKIE_AUTH_NONCE))
                .and(warp::cookie(COOKIE_AUTH_CSRF))
                .and(warp::query())
                .and_then(handle_auth_callback)
        };

        login_route.or(callback_route)
    };

    let frontend_files = warp::fs::dir(frontend_dir);

    let index = warp::fs::file(index_file);

    let routes = api_routes.or(ws).or(frontend_files).or(index);

    println!("Starting server on http://0.0.0.0:9000");
    warp::serve(routes).run(([0, 0, 0, 0], 9000)).await;
}

async fn ws_handler(
    name: String,
    ws: Ws,
    server: Arc<Server>,
    token: String,
) -> Result<impl Reply, Rejection> {
    let token: UserToken = token
        .verify_with_key(&server.signing_key)
        .map_err(|_| warp::reject())?;

    println!(
        "[INFO] User {} <{}> connected to document {}",
        token.display_name, token.email, name
    );
    let doc_id = format!("{}/{}", token.email, name);
    let connection = server.get_or_create_doc(doc_id).await;
    Ok(ws.on_upgrade(move |socket| peer(token, name, socket, connection)))
}

async fn peer(token: UserToken, doc_name: String, ws: WebSocket, connection: Arc<Connection>) {
    let (sink, stream) = ws.split();
    let sink = Arc::new(Mutex::new(WarpSink::from(sink)));
    let stream = WarpStream::from(stream);
    let sub = connection.bcast.subscribe(sink, stream);
    let _ = sub.completed().await;
    println!(
        "[INFO] User {} <{}> disconnected from document {}",
        token.display_name, token.email, doc_name
    );
}
