import { useState } from "react";
import { Button } from "~/components/ui/button";
import { Dialog, DialogClose, DialogContent, DialogDescription, DialogFooter, DialogHeader, DialogTitle, DialogTrigger } from "~/components/ui/dialog";
import { Input } from "~/components/ui/input";
import { Label } from "~/components/ui/label";
import { ColorPicker } from "~/components/form/color_picker";
import type { ColorName } from "~/lib/color";
import { IconPicker } from "~/components/form/icon_picker";
import type { IconName } from "~/lib/icon";
import { CollectionPicker } from "~/components/form/collection_picker";
import { type CollectionId, createNote, type NoteType } from "~/lib/metadata";
import { useNotesMetadata } from "~/hooks/use-metadata";
import { RadioGroup, RadioGroupItem } from "~/components/ui/radio-group";
import { useNavigate } from "@tanstack/react-router";

type NewNoteDialogProps = {
    children: React.ReactNode;
}

export function NewNoteDialog(props: NewNoteDialogProps) {
    const noteMetadata = useNotesMetadata();

    const [name, setName] = useState("");
    const [icon, setIcon] = useState<IconName>();
    const [collection, setCollection] = useState<CollectionId | undefined>();
    const [primaryColor, setPrimaryColor] = useState<ColorName | undefined>();
    const [secondaryColor, setSecondaryColor] = useState<ColorName | undefined>();
    const [type, setType] = useState<NoteType>("text");

    const navigate = useNavigate();
    function submit() {
        const newNote = createNote(noteMetadata, {
            name,
            icon,
            collectionId: collection,
            primaryColor,
            secondaryColor,
            type,
        });

        setName("");
        setIcon(undefined);
        setCollection(undefined);
        setPrimaryColor(undefined);
        setSecondaryColor(undefined);
        setType("text");

        navigate({ to: "/app/note/$id", params: { id: newNote.get("id") } });
    }

    return (
        <Dialog>
            <DialogTrigger asChild>
                {props.children}
            </DialogTrigger>
            <DialogContent>
                <DialogHeader>
                    <DialogTitle>New note</DialogTitle>
                    <DialogDescription>
                        Settings for the new collection
                    </DialogDescription>
                </DialogHeader>
                <div className="flex flex-col gap-2 p-2">
                    <div className="flex flex-row gap-2">
                        <IconPicker onChange={setIcon} />
                        <div className="grid flex-grow items-center gap-1.5">
                            <Label htmlFor="name">Name</Label>
                            <Input
                                id="name"
                                className="flex-grow"
                                placeholder="Name..."
                                onChange={(name) => setName(name.target.value)}
                            />
                        </div>
                    </div>
                    <div className="grid flex-grow items-center gap-1.5">
                        <Label>Collection</Label>
                        <CollectionPicker onChange={setCollection} className="w-full flex-grow" />
                    </div>
                    <div className="flex flex-row gap-2">
                        <div className="grid flex-grow items-center gap-1.5">
                            <Label>Primary color</Label>
                            <ColorPicker withNone onChange={setPrimaryColor} className="w-full flex-grow" />
                        </div>
                        <div className="grid flex-grow items-center gap-1.5">
                            <Label>Secondary color</Label>
                            <ColorPicker withNone onChange={setSecondaryColor} className="w-full flex-grow" />
                        </div>
                    </div>
                    <div className="grid flex-grow items-center gap-1.5">
                        <Label>Type</Label>
                        <RadioGroup
                            onValueChange={setType as any}
                            defaultValue={type}
                            className="flex flex-row space-y-1 items-center"
                        >
                            <div className="flex items-center space-x-2">
                                <RadioGroupItem value="text" id="type-text" />
                                <Label htmlFor="type-text">Text</Label>
                            </div>
                            <div className="flex items-center space-x-2">
                                <RadioGroupItem value="canvas" id="type-canvas" />
                                <Label htmlFor="type-canvas">Canvas</Label>
                            </div>
                            <div className="flex items-center space-x-2">
                                <RadioGroupItem value="entity" id="type-entity" />
                                <Label htmlFor="type-entity">Entity</Label>
                            </div>
                        </RadioGroup>
                    </div>
                </div>
                <DialogFooter>
                    <DialogClose asChild>
                        <Button
                            onClick={() => submit()}
                        >
                            Create
                        </Button>
                    </DialogClose>
                </DialogFooter>
            </DialogContent>
        </Dialog>
    );
}
