import * as Y from "yjs";
import { useCallback, useEffect } from "react";
import { Dialog, DialogContent, DialogDescription, DialogHeader, DialogTitle, DialogTrigger } from "~/components/ui/dialog";
import { Input } from "~/components/ui/input";
import type { NoteMetadata, NoteProperty } from "~/lib/metadata";
import { createNoteProperty, type PropertyType } from "~/lib/property";
import { useCollection } from "~/hooks/use-metadata";

type NotePropertiesDialogProps = {
    noteMetadata: NoteMetadata;
    children: React.ReactNode;
};
export function NotePropertiesDialog(props: NotePropertiesDialogProps) {
    const { noteMetadata } = props;

    if (!noteMetadata) {
        return props.children;
    }

    return (
        <Dialog>
            <DialogTrigger asChild>
                {props.children}
            </DialogTrigger>
            <DialogContent className="max-w-[800px]">
                <DialogHeader>
                    <DialogTitle>{noteMetadata.get("title")} properties</DialogTitle>
                    <DialogDescription>
                        Properties for the {noteMetadata.get("title")} note
                    </DialogDescription>
                </DialogHeader>
                <div className="h-[600px] overflow-scroll flex flex-col gap-4">
                    {noteMetadata.has("collectionId")
                        ? <PropertiesSection noteMetadata={noteMetadata} />
                        : <span>Only notes that belong to a collection can have properties</span>
                    }
                </div>
            </DialogContent>
        </Dialog>
    );
}

function PropertiesSection(props: { noteMetadata: NoteMetadata }) {
    const { noteMetadata } = props;
    const collection = useCollection(noteMetadata.get("collectionId"));

    useEffect(() => {
        if (!noteMetadata.has("properties")) {
            noteMetadata.set("properties", new Y.Array() as any)
        }
    }, [noteMetadata])

    const noteProperties = noteMetadata.get("properties");
    if (!noteProperties) {
        return null;
    }

    if (!collection) {
        return null;
    }

    const collectionProps = collection.get("properties")?.toArray();
    if (!collectionProps) {
        return null;
    }



    return (
        <div className="flex flex-col gap-2 p-2">
            {collectionProps.map((property) => {
                const propertyId = property.get("id");
                const noteProperty = noteProperties
                    ?.toArray()
                    .find(it => it.get("propertyId") == propertyId)
                return <PropertyItem key={propertyId}
                    noteProperty={noteProperty}
                    name={property.get("name")}
                    type={property.get("type")}
                    create={() => createNoteProperty(noteProperties, { propertyId })}
                />
            })}
        </div>
    );
}


type PropertyItemProps = {
    noteProperty?: NoteProperty;
    name: string;
    type: PropertyType;
    create: () => NoteProperty;
}
function PropertyItem(props: PropertyItemProps) {
    const handleChange = useCallback((value: string) => {
        if (props.noteProperty) {
            props.noteProperty.set("value", value);
        } else {
            props.create().set("value", value);
        }
    }, [props.noteProperty, props.create]);

    return (
        <div className="flex flex-row gap-2 items-center">
            <span className="w-[200px]">{props.name}</span>
            <Input
                className="flex-grow"
                defaultValue={props.noteProperty?.get("value")}
                placeholder={`${props.name}...`}
                onChange={(value) => handleChange(value.currentTarget.value)}
            />
        </div>
    );
}
