import { mdiMagnify } from "@mdi/js";
import Icon from "@mdi/react";
import { Link } from "@tanstack/react-router";
import { Command } from "cmdk";
import { colors } from "~/lib/color";
import { icons } from "~/lib/icon";
import type { NoteMetadata } from "~/lib/metadata";

type NotesGridProps = {
    notes?: NoteMetadata[];
    allUnpinned?: boolean;
};
export function NotesGrid(props: NotesGridProps) {
    const allUnpinned = props.allUnpinned ?? false;

    const pinnedNotes = props.notes?.filter(it => !allUnpinned && it.get("pinned"))?.map((note) => <DetailedNote key={note.get("id")} note={note} />);
    const normalNotes = props.notes?.filter(it => allUnpinned || !it.get("pinned"))?.map((note) => <NormalNote key={note.get("id")} note={note} />);

    return (
        <Command className="p-4 flex flex-col gap-4 items-stretch">
            <div className="bg-card rounded shadow flex flex-row items-center w-full max-w-[768px] mx-auto">
                <div className="p-2 flex justify-center items-center">
                    <Icon path={mdiMagnify} size={1.5} />
                </div>
                <Command.Input asChild>
                    <input type="text" placeholder="Search..." className="flex-grow bg-transparent py-4 px-2 outline-none" />
                </Command.Input>

            </div>
            <Command.List className="[&>div]:flex [&>div]:flex-col [&>div]:gap-4">
                <Command.Empty>No notes found</Command.Empty>
                <Command.Group
                    value="pinned"
                    className="[&>div]:grid [&>div]:w-full [&>div]:gap-4 [&>div]:grid-cols-[repeat(auto-fill,minmax(500px,1fr))]"
                >
                    {pinnedNotes}
                </Command.Group>

                <Command.Group
                    value="unpinned"
                    className="[&>div]:grid [&>div]:w-full [&>div]:gap-4 [&>div]:grid-cols-[repeat(auto-fill,minmax(300px,1fr))]"
                >
                    {normalNotes}
                </Command.Group>
            </Command.List>
        </Command>
    );
}

type DetailedNoteProps = {
    note: NoteMetadata;
}

function DetailedNote(props: DetailedNoteProps) {
    const { note } = props;

    const primaryColorName = note.get("primaryColor")
    const secondaryColorName = note.get("secondaryColor")

    const primaryColor = primaryColorName ? colors[primaryColorName] : colors.white;
    const secondaryColor = secondaryColorName ? colors[secondaryColorName] : primaryColor;
    const icon = note.get("icon") ? icons[note.get("icon")] : undefined;

    return (
        <Command.Item asChild value={note.get("title")}>
            <Link to="/app/note/$id" params={{ id: note.get("id") }} className="bg-card shadow shadow-black rounded overflow-hidden relative">
                <div className="w-full h-[100px]" style={{
                    background: `linear-gradient(135deg, ${primaryColor.base}, ${secondaryColor.base})`
                }} />
                {icon && <div className="inline-block bg-zinc-900 absolute top-[58px] left-4 rounded-xl p-2">
                    <Icon path={icon.path} size={2} color={primaryColor.base} />
                </div>}
                <div className="px-4 pt-2 pb-4 mt-[22px]">
                    <h1 className="ml-2 font-bold text-xl">{note.get("title")}</h1>
                </div>
            </Link>
        </Command.Item>
    );
}

type NormalNoteProps = {
    note: NoteMetadata;
};
function NormalNote(props: NormalNoteProps) {
    const { note } = props;

    const colorName = note.get("primaryColor");

    const color = colorName ? colors[colorName] : colors.white;
    const icon = note.get("icon") ? icons[note.get("icon")] : undefined;

    return (
        <Command.Item asChild value={note.get("title")}>
            <Link to="/app/note/$id" params={{ id: note.get("id") }} className="bg-card shadow rounded overflow-hidden flex flex-row">
                <div className="flex-shrink-0 w-1 h-full" style={{ backgroundColor: color.base }} />
                <div className="flex flex-col gap-2 px-4 py-2 justify-center">
                    <div className="flex flex-row gap-2 items-center">
                        {icon && <Icon path={icon.path} size={1.5} color={color.base} />}
                        <h1 className="font-bold text-xl">{note.get("title")}</h1>
                    </div>
                </div>
            </Link>
        </Command.Item>
    );
}
