import { $createParagraphNode, DecoratorNode, ElementNode, type LexicalNode, ParagraphNode, type RangeSelection, type SerializedElementNode, type SerializedLexicalNode, type Spread } from "lexical";
import { FocusableNode } from "./focusable_node";
import type { ReactNode } from "react";
import { type TaskType, taskTypes } from "../plugins/task_plugin";
import { TaskIcon } from "~/components/editor/task_icon";
import type { TaskLabel } from "../serialized_editor_content";

type SerializedTaskNode = Spread<SerializedElementNode, {
    taskType: TaskLabel;
}>;

export class TaskNode extends FocusableNode {
    __taskType: TaskType;

    getTaskType() {
        const self = this.getLatest();
        return self.__taskType;
    }

    static getType() {
        return "task";
    }
    static clone(node: TaskNode) {
        return new TaskNode(node.__taskType, node.__key, node.__hasFocus);
    }

    constructor(taskType: TaskType, key?: string, hasFocus?: boolean) {
        super(key, hasFocus);
        this.__taskType = taskType;
    }

    createDOM(): HTMLElement {
        const element = super.createDOM();
        element.classList.add("task");
        element.setAttribute("data-task-type", this.__taskType.label);
        return element;
    }
    updateDOM(old: TaskNode, dom: HTMLElement): boolean {
        return super.updateDOM(old, dom);
    }
    insertNewAfter(_selection: RangeSelection, restoreSelection?: boolean): ParagraphNode {
        const newElement = $createParagraphNode();
        this.insertAfter(newElement, restoreSelection);
        return newElement;
    }
    exportJSON(): SerializedTaskNode {
        return {
            ...super.exportJSON(),
            type: "task",
            taskType: this.__taskType.label,
        };
    }
    static importJSON(serializedTaskNode: SerializedTaskNode) {
        const taskType = taskTypes.find(it => it.label == serializedTaskNode.taskType);
        // FIXME: Deal with task type not existing gracefully. Uncreate the node?
        return $createTaskNode(taskType!);
    }
}

export function $createTaskNode(taskType: TaskType) {
    return new TaskNode(taskType);
}

export function $isTaskNode(node?: LexicalNode | null): node is TaskNode {
    return node instanceof TaskNode;
}

export class TaskMarkerNode extends ElementNode {
    static getType() {
        return "task-marker";
    }
    static clone(node: TaskMarkerNode) {
        return new TaskMarkerNode(node.__key);
    }
    isInline(): boolean {
        return true;
    }
    createDOM(): HTMLElement {
        const element = document.createElement("span");
        element.classList.add("task-marker");
        return element;
    }
    updateDOM(): boolean {
        return false;
    }
    exportJSON() {
        return {
            ...super.exportJSON(),
            type: "task-marker",
        };
    }
    static importJSON() {
        return $createTaskMarkerNode();
    }
}

export function $createTaskMarkerNode() {
    return new TaskMarkerNode();
}

export function $isTaskMarkerNode(node?: LexicalNode | null): node is TaskMarkerNode {
    return node instanceof TaskMarkerNode;
}


type SerializedTaskIconNode = Spread<SerializedLexicalNode, {
    taskType: TaskLabel;
}>;

export class TaskIconNode extends DecoratorNode<ReactNode> {
    __taskType: TaskType;

    getTaskType() {
        const self = this.getLatest();
        return self.__taskType;
    }

    static getType() {
        return "task-icon";
    }
    static clone(node: TaskIconNode) {
        return new TaskIconNode(node.__taskType, node.__key);
    }

    constructor(taskType: TaskType, key?: string) {
        super(key);
        this.__taskType = taskType;
    }

    createDOM(): HTMLElement {
        return document.createElement("span");
    }
    updateDOM(): boolean {
        return false;
    }
    decorate(): ReactNode {
        return (<TaskIcon icon={this.__taskType.icon?.path} />);
    }
    exportJSON(): SerializedTaskIconNode {
        return {
            ...super.exportJSON(),
            type: "task-icon",
            taskType: this.__taskType.label,
        };
    }
    static importJSON(serializedTaskIconNode: SerializedTaskIconNode) {
        const taskType = taskTypes.find(it => it.label == serializedTaskIconNode.taskType);
        // FIXME: Deal with task type not existing gracefully. Uncreate the node?
        return $createTaskIconNode(taskType!);
    }
}

export function $createTaskIconNode(taskType: TaskType) {
    return new TaskIconNode(taskType);
}

export function $isTaskIconNode(node?: LexicalNode | null): node is TaskIconNode {
    return node instanceof TaskIconNode;
}
