import { useLexicalComposerContext } from "@lexical/react/LexicalComposerContext";
import { $createParagraphNode, $isParagraphNode, $isTextNode, SELECTION_CHANGE_COMMAND, TextNode, type LexicalEditor, type LexicalNode } from "lexical";
import { useEffect } from "react";
import { $createHeaderMarkerNode, $createHeaderNode, $isHeaderMarkerNode, $isHeaderNode, HeaderMarkerNode, HeaderNode } from "../nodes/header_node";
import { mergeRegister } from "@lexical/utils";
import { findBlockNode } from "../editor_utils";

const HEADER_REGEX = /^#+ /;

export function HeaderPlugin() {
    const [editor] = useLexicalComposerContext();

    useEffect(() => mergeRegister(
        // Create a header node if the text node matches the HEADER_REGEX
        editor.registerNodeTransform(TextNode, createHeaderTransform(editor)),
        editor.registerNodeTransform(TextNode, ensureHeaderValidTransform),
        editor.registerNodeTransform(HeaderNode, ensureHeaderValidTransform),
        // Remove header markers if they aren't in a header
        editor.registerNodeTransform(HeaderMarkerNode, (node) => {
            const headerNode = node.getParent();
            if ($isHeaderNode(headerNode)) {
                return;
            }

            node.getChildren().reverse().forEach(child => node.insertAfter(child));
            node.remove();
        }),
    ));

    return null;
}

const createHeaderTransform = (editor: LexicalEditor) => (textNode: TextNode) => {
    const prevNode = textNode.getPreviousSibling();
    if (prevNode) return;
    const paragraphNode = textNode.getParent();
    if (!$isParagraphNode(paragraphNode)) return;

    const content = textNode.getTextContent();
    const regexMatch = content.match(HEADER_REGEX);
    if (!regexMatch) return;

    const children = paragraphNode.getChildren();

    const firstTextNode = children[0];
    if (!$isTextNode(firstTextNode)) return;

    const markerLength = regexMatch[0].length;
    const textNodes = firstTextNode.splitText(markerLength);

    const headerMarkerContent = textNodes[0];
    if (!headerMarkerContent) return;

    const headerNode = $createHeaderNode(markerLength - 1);
    const headerMarkerNode = $createHeaderMarkerNode();

    headerMarkerNode.append(headerMarkerContent);

    headerNode.append(headerMarkerNode);
    headerNode.append(...textNodes.slice(1));
    headerNode.append(...children.slice(1));

    paragraphNode.replace(headerNode, true);

    editor.dispatchCommand(SELECTION_CHANGE_COMMAND, undefined);
}

const ensureHeaderValidTransform = (node: LexicalNode) => {
    const headerNode = findBlockNode(node);
    if (!$isHeaderNode(headerNode)) {
        return;
    }

    const markerNode = headerNode.getFirstChild();
    const markerContent = markerNode?.getTextContent();
    if ($isHeaderMarkerNode(markerNode)
        && markerContent?.match(HEADER_REGEX)
        && markerContent.length == headerNode.getLevel() + 1) {
        return
    }

    headerNode.replace($createParagraphNode(), true);
}
