import { useLexicalComposerContext } from "@lexical/react/LexicalComposerContext";
import { mergeRegister } from "@lexical/utils";
import { SELECTION_CHANGE_COMMAND, TextNode } from "lexical";
import { useEffect } from "react";
import { $createLinkIconNode, $createLinkMarkerNode, $createLinkNode, $createLinkUrlNode, $isLinkIconNode, $isLinkMarkerNode, $isLinkNode, $isLinkUrlNode, LinkIconNode, LinkMarkerNode, LinkNode, LinkUrlNode } from "../nodes/link_node";

const LINK_REGEX = /\[(.+)\]\((.*)\)/;

export function LinkPlugin() {
    const [editor] = useLexicalComposerContext();

    useEffect(() => mergeRegister(
        // Create link node
        editor.registerNodeTransform(TextNode, (node) => {
            const content = node.getTextContent();
            const matches = content.match(LINK_REGEX);
            if (!matches) return;

            const labelLength = matches[1]!.length;
            const urlLength = matches[2]!.length;

            // start
            //  | labelStart
            //  | | labelEnd
            //  | |     | middle
            //  | |     | | urlStart
            //  | |     | | | urlEnd
            //  | |     | | |   | end   Index of the segment
            //  | |     | | |   | |     in the textNodes array
            // 0|1|2    |3|4|5  |6|7 <--|
            //  |[|label|]|(|url|)|
            const start = content.indexOf(matches[0]);
            const labelStart = start + 1;
            const labelEnd = labelStart + labelLength;
            const middle = labelEnd + 1;
            const urlStart = middle + 1;
            const urlEnd = urlStart + urlLength;
            const end = urlEnd + 1;

            const textNodes = node.splitText(start, labelStart, labelEnd, middle, urlStart, urlEnd, end);
            const labelStartIndex = start > 0 ? 1 : 0;
            const labelIndex = labelStartIndex + 1;
            const labelEndIndex = labelIndex + 1;
            const urlStartIndex = labelEndIndex + 1;
            const urlIndex = urlStartIndex + 1;
            const urlEndIndex = urlIndex + 1;

            const url = matches[2]!;

            const linkNode = $createLinkNode(url);
            textNodes[labelStartIndex]!.insertBefore(linkNode);

            const labelStartNode = $createLinkMarkerNode();
            labelStartNode.append(textNodes[labelStartIndex]!);
            const labelEndNode = $createLinkMarkerNode();
            labelEndNode.append(textNodes[labelEndIndex]!);
            const urlStartNode = $createLinkMarkerNode();
            urlStartNode.append(textNodes[urlStartIndex]!);
            const urlNode = $createLinkUrlNode();
            urlNode.append(textNodes[urlIndex]!);
            const urlEndNode = $createLinkMarkerNode();
            urlEndNode.append(textNodes[urlEndIndex]!);
            const iconNode = $createLinkIconNode(url);

            linkNode.append(labelStartNode, textNodes[labelIndex]!, labelEndNode, urlStartNode, urlNode, urlEndNode, iconNode);

            editor.dispatchCommand(SELECTION_CHANGE_COMMAND, undefined);
        }),
        // Remove dangling link marker nodes and verify their content
        editor.registerNodeTransform(LinkMarkerNode, (node) => {
            const linkNode = node.getParent();
            if ($isLinkNode(linkNode)) {
                const content = node.getTextContent();
                const index = node.getIndexWithinParent();
                const parentChildCount = linkNode.getChildren().length;

                if (index === 0 && content === "[") return;
                if (index === parentChildCount - 5 && content === "]") return;
                if (index === parentChildCount - 4 && content === "(") return;
                if (index === parentChildCount - 2 && content === ")") return;
            }

            node.getChildren().reverse().forEach((child) => node.insertAfter(child));
            node.remove();
        }),
        // Remove dangling link url nodes
        editor.registerNodeTransform(LinkUrlNode, (node) => {
            const linkNode = node.getParent();
            if ($isLinkNode(linkNode)) {
                const url = linkNode.getUrl();
                const content = node.getTextContent();
                if (url === content) return;
            }

            node.getChildren().reverse().forEach((child) => node.insertAfter(child));
            node.remove();
        }),
        // Remove dangling link icon nodes
        editor.registerNodeTransform(LinkIconNode, (node) => {
            const linkNode = node.getParent();
            if ($isLinkNode(linkNode)) {
                const url = linkNode.getUrl();
                const iconUrl = node.getUrl();
                if (url === iconUrl) return;
            }

            node.remove();
        }),
        // Remove links with missing markers
        editor.registerNodeTransform(LinkNode, (node) => {
            const children = node.getChildren();
            const startLabel = children.at(0);
            const endLabel = children.at(-5);
            const startUrl = children.at(-4);
            const url = children.at(-3);
            const endUrl = children.at(-2);
            const icon = children.at(-1);

            if (
                $isLinkMarkerNode(startLabel) &&
                $isLinkMarkerNode(endLabel) &&
                $isLinkMarkerNode(startUrl) &&
                $isLinkUrlNode(url) &&
                $isLinkMarkerNode(endUrl) &&
                $isLinkIconNode(icon)
            ) {
                return;
            }

            children.reverse().forEach((child) => node.insertAfter(child));
            node.remove();
        }),
    ))

    return null;
}
