import { useLexicalComposerContext } from "@lexical/react/LexicalComposerContext";
import { mergeRegister } from "@lexical/utils";
import { $createParagraphNode, $isParagraphNode, $isTextNode, type LexicalEditor, SELECTION_CHANGE_COMMAND, TextNode } from "lexical";
import { useEffect } from "react";
import { $createTaskMarkerNode, $createTaskNode, $createTaskIconNode, TaskMarkerNode, $isTaskNode, TaskIconNode, TaskNode, $isTaskIconNode, $isTaskMarkerNode } from "../nodes/task_node";
import { type Icon, icons } from "~/lib/icon";
import { type TaskLabel } from "../serialized_editor_content";

export type TaskType = {
    label: TaskLabel
    icon?: Icon
}

export const taskTypes: TaskType[] = [
    {
        label: "TODO",
        icon: undefined,
    },
    {
        label: "DOING",
        icon: icons.minus,
    },
    {
        label: "DONE",
        icon: icons.check,
    },
    {
        label: "IDEA",
        icon: icons.lightbulb,
    },
    {
        label: "DEADLINE",
        icon: icons.exclamation,
    },
]

export function TaskPlugin() {
    const [editor] = useLexicalComposerContext();

    useEffect(() => mergeRegister(
        // Create a task node if the text node matches the TASK_REGEX
        ...taskTypes.map(({ label }) =>
            editor.registerNodeTransform(TextNode, createTaskNode(label, editor))
        ),
        // Remove task marker nodes not inside a task node or with the wrong content
        editor.registerNodeTransform(TaskMarkerNode, (node) => {
            const taskNode = node.getParent();
            if ($isTaskNode(taskNode)) {
                const taskType = taskNode.getTaskType();
                const content = node.getTextContent();
                if (content === taskType.label) return;
            }

            node.getChildren().reverse().forEach(child => node.insertAfter(child));
            node.remove();
        }),
        // Remove task icon nodes not inside a task node
        editor.registerNodeTransform(TaskIconNode, (node) => {
            const taskNode = node.getParent();
            if ($isTaskNode(taskNode)) {
                const taskType = taskNode.getTaskType();
                if (node.getTaskType() === taskType) return;
            }

            node.remove();
        }),
        // Remove task nodes without an icon or marker
        editor.registerNodeTransform(TaskNode, (node) => {
            const iconNode = node.getChildAtIndex(0);
            const markerNode = node.getChildAtIndex(1);

            if ($isTaskIconNode(iconNode) && $isTaskMarkerNode(markerNode)) return;

            node.replace($createParagraphNode(), true);
        }),
    ));

    return null;
}

const createTaskNode = (label: TaskLabel, editor: LexicalEditor) => (node: TextNode) => {
    const prevNode = node.getPreviousSibling();
    if (prevNode) return;
    const paragraphNode = node.getParent();
    if (!$isParagraphNode(paragraphNode)) return;

    const content = node.getTextContent();
    if (!content.startsWith(label + " ")) return;

    const children = paragraphNode.getChildren();

    const firstTextNode = children[0];
    if (!$isTextNode(firstTextNode)) return;

    const textNodes = firstTextNode.splitText(label.length);

    const todoMarkerContent = textNodes[0];
    if (!todoMarkerContent) return;

    const taskType = taskTypes.find(it => it.label == label);
    if (!taskType) return;

    const taskNode = $createTaskNode(taskType);
    const taskIconNode = $createTaskIconNode(taskType);
    const taskMarkerNode = $createTaskMarkerNode();

    taskMarkerNode.append(todoMarkerContent);

    taskNode.append(taskIconNode);
    taskNode.append(taskMarkerNode);
    taskNode.append(...textNodes.slice(1));
    taskNode.append(...children.slice(1));

    paragraphNode.replace(taskNode, true);
    editor.dispatchCommand(SELECTION_CHANGE_COMMAND, undefined);
}
