// TODO: This needs to be removed, but some exported types need to be moved first
import { $getRoot, $isLineBreakNode, $isParagraphNode, $isTextNode, type LexicalNode } from "lexical";
import { $isHeaderNode } from "./nodes/header_node";
import { $isTaskNode } from "./nodes/task_node";
import { $isFormattedTextNode } from "./nodes/formatted_text";
import { $isLinkNode } from "./nodes/link_node";
import { $isTermNode } from "./nodes/term_node";

export type SerializedEditorContent = {
    content: SerializedBlockNode[];
}

type SerializedTextContainingNode =
    SerializedTextNode
    | SerializedFormattedTextNode
    | SerializedLinebreakNode
    | SerializedLinkNode
    | SerializedTermNode;
type SerializedBlockNode =
    SerializedHeaderNode
    | SerializedParagraphNode
    | SerializedTaskNode;

type SerializedTextNode = {
    type: "text";
    content: string;
}

type SerializedLinebreakNode = {
    type: "linebreak";
}

type SerializedParagraphNode = {
    type: "paragraph";
    content: SerializedTextContainingNode[];
}

type SerializedHeaderNode = {
    type: "header";
    level: number;
    content: SerializedTextContainingNode[];
}

export type TextFormat = "bold" | "italic" | "underline" | "strikethrough" | "code";
export const textFormats: Record<TextFormat, string> = {
    bold: "**",
    italic: "//",
    underline: "__",
    strikethrough: "~~",
    code: "`",
};
type SerializedFormattedTextNode = {
    type: "formatted_text";
    format: TextFormat;
    content: SerializedTextContainingNode[];
}

type SerializedLinkNode = {
    type: "link";
    label: SerializedTextContainingNode[];
    url: string;
}

type SerializedTermNode = {
    type: "term";
    term: string;
}


export type TaskLabel = "TODO" | "DOING" | "DONE" | "IDEA" | "DEADLINE";
type SerializedTaskNode = {
    type: "task";
    label: TaskLabel;
    content: SerializedTextContainingNode[];

}

export function $serializeEditorContent(): SerializedEditorContent {
    return {
        content: $getRoot().getChildren().map(serializeBlockNode),
    };
}

function serializeBlockNode(node: LexicalNode): SerializedBlockNode {
    switch (true) {
        case $isParagraphNode(node): return {
            type: "paragraph",
            content: node.getChildren().map(serializeTextContainingNode),
        }
        case $isHeaderNode(node): return {
            type: "header",
            level: node.getLevel(),
            // Slice(1) to remove the marker node
            content: node.getChildren().slice(1).map(serializeTextContainingNode),
        }
        case $isTaskNode(node): return {
            type: "task",
            label: node.getTaskType().label,
            // Slice(2) to remove the marker nodes
            content: node.getChildren().slice(2).map(serializeTextContainingNode),
        }
        default: throw new Error(`Unknown block node type: ${node.getType()}`);
    }
}

function serializeTextContainingNode(node: LexicalNode): SerializedTextContainingNode {
    switch (true) {
        case $isTextNode(node): return {
            type: "text",
            content: node.getTextContent(),
        }
        case $isFormattedTextNode(node): return {
            type: "formatted_text",
            format: node.getStyle(),
            content: node.getChildren().slice(1, -1).map(serializeTextContainingNode),
        }
        case $isLineBreakNode(node): return {
            type: "linebreak",
        }
        case $isLinkNode(node): return {
            type: "link",
            label: node.getChildren().slice(1, -5).map(serializeTextContainingNode),
            url: node.getUrl(),

        }
        case $isTermNode(node): return {
            type: "term",
            term: node.getTerm(),
        }
        default: throw new Error(`Unknown text containing node type: ${node.getType()}`);
    }
}

export function deserializeEditorContent(serialized: SerializedEditorContent) {
    return {
        root: {
            type: "root",
            children: serialized.content.map(deserializeBlockNode),
        }
    };
}

function deserializeBlockNode(serialized: SerializedBlockNode) {
    switch (serialized.type) {
        case "header": return {
            type: "header",
            level: serialized.level,
            children: [
                {
                    type: "header-marker",
                    children: [createTextNode(`${"#".repeat(serialized.level)} `)]
                },
                ...serialized.content.map(deserializeTextContainingNode)
            ],
        }
        case "paragraph": return {
            type: "paragraph",
            children: [
                ...serialized.content.map(deserializeTextContainingNode)
            ],
        }
        case "task": return {
            type: "task",
            taskType: serialized.label,
            children: [
                {
                    type: "task-icon",
                    taskType: serialized.label,
                },
                {
                    type: "task-marker",
                    children: [createTextNode(`${serialized.label} `)]
                },
                ...serialized.content.map(deserializeTextContainingNode)
            ],
        }
    }
}

function deserializeTextContainingNode(serialized: SerializedTextContainingNode): any {
    switch (serialized.type) {
        case "text": return createTextNode(serialized.content);
        case "formatted_text": return {
            type: "formatted-text",
            style: serialized.format,
            children: [
                {
                    type: "formatted-text-marker",
                    children: [createTextNode(textFormats[serialized.format])]
                },
                ...serialized.content.map(deserializeTextContainingNode),
                {
                    type: "formatted-text-marker",
                    children: [createTextNode(textFormats[serialized.format])]
                },
            ],
        }
        case "link": return {
            type: "link",
            children: [
                {
                    type: "link-marker",
                    children: [createTextNode("[")],
                },
                ...serialized.label.map(deserializeTextContainingNode),
                {
                    type: "link-marker",
                    children: [createTextNode("]")],
                },
                {
                    type: "link-marker",
                    children: [createTextNode("(")],
                },
                {
                    type: "link-url",
                    children: [createTextNode(serialized.url)],
                },
                {
                    type: "link-marker",
                    children: [createTextNode(")")],
                },
                {
                    type: "link-icon",
                    url: serialized.url,
                }
            ],
        };
        case "term": return {
            type: "term",
            term: serialized.term,
            children: [
                {
                    type: "term-icon",
                    term: serialized.term,
                },
                {
                    type: "term-marker",
                    children: [createTextNode("[[")]
                },
                createTextNode(serialized.term),
                {
                    type: "term-marker",
                    children: [createTextNode("]]")]
                },

            ],
        }
        case "linebreak": return { type: "linebreak" };
    }
}

function createTextNode(content: string) {
    return {
        type: "text",
        text: content,

        version: 1,
        detail: 0,
        format: 0,
        mode: "normal",
        style: "",
    }
}
