import { createContext, useContext } from "react";
import Cookies from "js-cookie";
import { useNavigate } from "@tanstack/react-router";
import { useRedraw } from "./use-redraw";

type AuthContextType = {
    jwt?: {
        display_name: string;
        email: string;

        iat: number,
        jti: string,
    },
    signOut: () => void;
};

const AUTH_TOKEN_COOKIE_NAME = "knotes_auth_token";

const AuthContext = createContext<AuthContextType | null>(null);

export function AuthProvider(props: { children: React.ReactNode }) {
    const redraw = useRedraw();
    const navigate = useNavigate();
    const authToken = Cookies.get(AUTH_TOKEN_COOKIE_NAME);
    const jwt: AuthContextType["jwt"] = authToken ? parseJwt(authToken) : null;

    function signOut() {
        // Remove the auth token
        Cookies.remove(AUTH_TOKEN_COOKIE_NAME);

        // Redirect to the landing page
        navigate({ to: "/" }).then(() => {
            // Force refresh the auth token from the cookie
            redraw();

            // After navigation completes, delete all locally stored data.
            // We do this because otherwise the providers might still be mounted
            // and might resync the data before they are unmounted.
            indexedDB.databases().then((dbs) => {
                dbs.forEach((db) => {
                    db.name && indexedDB.deleteDatabase(db.name);
                });
            });
        });
    }

    return (
        <AuthContext.Provider value={{ jwt, signOut }}>
            {props.children}
        </AuthContext.Provider>
    );
}

export function useAuth() {
    const userData = useContext(AuthContext);

    if (!userData) {
        throw new Error("useAuth must be used within an AuthProvider");
    }

    return {
        isAuthenticated: !!userData?.jwt,
        data: userData.jwt,
        signOut: userData.signOut,
    };
}

function parseJwt(token: string) {
    var base64Url = token.split('.')[1];
    var base64 = base64Url.replace(/-/g, '+').replace(/_/g, '/');
    var jsonPayload = decodeURIComponent(window.atob(base64).split('').map(function(c) {
        return '%' + ('00' + c.charCodeAt(0).toString(16)).slice(-2);
    }).join(''));

    return JSON.parse(jsonPayload);
}
