import React, { useCallback, useContext, useLayoutEffect, useState } from "react";

type Theme = "system" | "light" | "dark";

const ThemeContex = React.createContext<{ theme: Theme, effectiveTheme: "light" | "dark", setTheme: (theme: Theme) => void }>({
    theme: "system",
    effectiveTheme: "light",
    setTheme: () => { }
});

export function ThemeProvider(props: { children: React.ReactNode }) {
    const [localTheme, setLocalTheme] = useState<Theme>(() => localStorage.getItem("theme") as Theme || "system");
    const systemTheme = window.matchMedia("(prefers-color-scheme: dark)").matches ? "dark" : "light";

    const effectiveTheme = localTheme === "system" ? systemTheme : localTheme;

    useLayoutEffect(() => {
        document.documentElement.setAttribute("data-theme", effectiveTheme)
    }, [effectiveTheme])

    const setTheme = useCallback((newTheme: Theme) => {
        setLocalTheme(newTheme);
        localStorage.setItem("theme", newTheme);
    }, []);

    return (
        <ThemeContex.Provider value={{
            theme: localTheme,
            effectiveTheme,
            setTheme
        }}>
            {props.children}
        </ThemeContex.Provider>
    );
}

export function useTheme() {
    const { theme, setTheme } = useContext(ThemeContex);

    return [theme, setTheme] as const;
}

export function useEffectiveTheme() {
    const { effectiveTheme } = useContext(ThemeContex);

    return effectiveTheme;
}
